# -*- coding: utf-8 -*-
from glob import glob
import sys
import os
import argparse
import subprocess
import pipes
from collections import namedtuple
from distutils.dir_util import copy_tree

from .version import kima_version
from kima.paths import kima_dir


def usage():
    u = "kima-template [DIRECTORY]\n"
    u += "Create a kima template in DIRECTORY (the current directory by default).\n"
    return u


def _parse_args():
    parser = argparse.ArgumentParser(usage=usage())
    parser.add_argument('DIRECTORY', nargs='?', default=os.getcwd())
    parser.add_argument('-e', '--example', type=str, choices=['14Her'],
                        help='Start from an example')
    parser.add_argument('--dace', type=str, metavar='STAR',
                        help='download .rdb files for STAR from DACE')
    parser.add_argument('--version', action='store_true',
                        help='show version and exit')
    parser.add_argument('--debug', action='store_true',
                        help='be more verbose, for debugging')
    return parser.parse_args()


def exists_remote(host, path, verbose=False):
    if verbose:
        print('[mkdir_remote] checking if "%s" exists in %s' % (path, host))
    return subprocess.call(['ssh', host, 'test -e ' + pipes.quote(path)]) == 0


def mkdir_remote(host, path, verbose=False):
    if verbose:
        print('[mkdir_remote] creating directory "%s" in %s' % (path, host))
    subprocess.check_call(['ssh', host, 'mkdir ' + pipes.quote(path)])


def write_makefile(directory):
    from os.path import join
    import sysconfig
    from textwrap import dedent

    lib = sysconfig.get_config_vars('EXT_SUFFIX')[0].replace('.so', '.a')
    kima_lib = f'libkima{lib}'
    dnest4_lib = f'libdnest4{lib}'
    # print(dnest4_lib, kima_lib)

    make = f"""
    KIMA_DIR = {kima_dir}

    DNEST4_PATH = $(KIMA_DIR)/vendor/DNest4/code
    EIGEN_PATH = $(KIMA_DIR)/vendor/eigen
    LOADTXT_PATH = $(KIMA_DIR)/vendor/cpp-loadtxt/src
    INCLUDES = -I$(KIMA_DIR) -I$(DNEST4_PATH) -I$(EIGEN_PATH) -I$(LOADTXT_PATH)

    LIBS = -L$(KIMA_DIR) -l:{dnest4_lib}

    CXXFLAGS = -pthread -std=c++17 -O3 -DNDEBUG -DEIGEN_MPL2_ONLY

    kima: kima_setup.cpp
    \t@$(CXX) -o $@ $< $(CXXFLAGS) $(KIMA_DIR)/{kima_lib} $(INCLUDES) $(LIBS)

    clean:
    \trm -f kima kima_setup.o
    """
    with open(join(directory, 'Makefile'), 'w') as f:
        print(dedent(make), file=f)


def main(args=None, stopIfNoReplace=True):
    if args is None:
        args = _parse_args()
    if isinstance(args, str):
        Args = namedtuple('Args', 'version debug DIRECTORY')
        args = Args(version=False, debug=False, DIRECTORY=args)

    if args.version:
        # same as kima
        print('kima (kima-template script)', kima_version)
        return

    dst = args.DIRECTORY

    if args.example:
        src = os.path.join(kima_dir, 'examples', args.example)
        templatefiles = glob(os.path.join(src, '*'))
    else:
        thisdir = os.path.dirname(os.path.realpath(__file__))
        src = os.path.join(thisdir, 'template')
        templatefiles = ['kima_setup.cpp', 'OPTIONS']

    # by default, replace directory
    replace = True

    if ':' in dst and '@' in dst:  # directory is in the server
        server = True
        host, directory = dst.split(':')
        # user, host = server.split('@')
        dst = directory
        print('Populating directory %s (on %s)' % (dst, host))
        # print('with kima template files: '
        #       ', '.join([os.path.basename(f) for f in templatefiles]))

        answer = input('--> Continue? (y/N) ')
        if answer.lower() != 'y':
            print('doing nothing!')
            return
    else:
        server = False

        if os.path.exists(dst):
            # is there a kima_setup.cpp inside the directory?
            if os.path.exists(os.path.join(dst, 'kima_setup.cpp')):
                msg = 'Directory "%s" exists and contains a kima_setup file. ' % dst
                msg += 'Replace? (Y/n) '
                print(msg, end='')
                answer = input().lower()
                if answer == 'n':
                    replace = False
                    if stopIfNoReplace:
                        print('Doing nothing')
                        sys.exit(0)
                else:
                    replace = True

            # cwd obviously exists, but make sure not to overwrite other directories
            elif dst != os.getcwd():
                print('Directory "%s" exists. Continue? (Y/n) ' % dst, end='')
                answer = input().lower()
                if answer == 'n':
                    print('Doing nothing')
                    sys.exit(0)
                replace = True

        if replace:
            print(f'Populating directory "{dst}"')
            print('with kima template files:', end=' ')
            print(', '.join([os.path.basename(f) for f in templatefiles]))

    if server:
        pass
    #     if not exists_remote(host, dst, verbose=args.debug):
    #         mkdir_remote(host, dst, verbose=args.debug)
    #     for f in templatefiles:
    #         # -q -o LogLevel=QUIET
    #         cmd = 'scp %s %s' % (f, ':'.join([host, dst]))
    #         # print(cmd)
    #         subprocess.check_call(cmd.split())
    else:
        if replace:
            copy_tree(src, dst)
        write_makefile(dst)

    # if server:
    #     with open(os.path.join(src, 'Makefile')) as f:
    #         m = f.read().format(kimadir='/home/jfaria/disk1/kima')
    #     with open('tempMakefile1234', 'w') as f:
    #         f.write(m)
    #     cmd = 'scp -q -o LogLevel=QUIET %s %s' % ('tempMakefile1234', host +
    #                                               ':' + dst + '/Makefile')
    #     # print(cmd)
    #     subprocess.check_call(cmd.split())
    #     os.remove('tempMakefile1234')
    # else:
    #     with open(os.path.join(src, 'Makefile')) as f:
    #         m = f.read().format(kimadir=kimadir)
    #     with open(os.path.join(dst, 'Makefile'), 'w') as f:
    #         f.write(m)

    if args.dace is not None:
        raise NotImplementedError


    #     from vera import DACE
    #     from .utils import chdir
    #     with chdir(dst):
    #         s = getattr(DACE, args.dace)
    #         print(s)

    # return


if __name__ == '__main__':
    main()



""" Small (but sometimes important) utility functions """

from datetime import datetime
import os
import re
import contextlib
import math
import configparser

import numpy as np
from scipy import stats
from astropy.units import R_sun, AU

import urepr
from loguniform import ModifiedLogUniform
from kumaraswamy import kumaraswamy

# CONSTANTS
mjup2mearth = 317.8284065946748       # 1 Jupiter mass in Earth masses
mjup2msun = 0.0009545942339693249     # 1 Jupiter mass in solar masses
mearth2msun = 3.0034893488507934e-06  # 1 Earth mass in solar masses


def get_kima_dir():
    here = os.path.dirname(os.path.dirname(__file__))
    return here


def read_model_setup(filename='kima_model_setup.txt'):
    setup = configparser.ConfigParser()
    setup.optionxform = str
    setup.read(filename)
    # if sys.version_info < (3, 0):
    #     setup = setup._sections
    #     # because we cheated, we need to cheat a bit more...
    #     setup['kima']['GP'] = setup['kima'].pop('gp')
    return setup


def read_datafile(datafile, skip):
    """
    Read data from `datafile` for multiple instruments.
    Can be str, in which case the 4th column is assumed to contain an integer
    identifier of the instrument.
    Or list, in which case each element will be one different filename
    containing three columns each.
    """
    if isinstance(datafile, list):
        data = np.empty((0, 3))
        obs = np.empty((0, ))
        for i, df in enumerate(datafile):
            d = np.loadtxt(df, usecols=(0, 1, 2), skiprows=skip, ndmin=2)
            data = np.append(data, d, axis=0)
            obs = np.append(obs, (i + 1) * np.ones((d.shape[0])))
        return data, obs
    else:
        data = np.loadtxt(datafile, usecols=(0, 1, 2), skiprows=skip)
        obs = np.loadtxt(datafile, usecols=(3, ), skiprows=skip, dtype=int)
        uobs = np.unique(obs)
        if uobs.min() > 0:
            uobs -= uobs.min()

        return data, obs


def read_datafile_rvfwhm(datafile, skip):
    """
    Read data from `datafile` for multiple instruments and RV-FWHM data.
    Can be str, in which case the 4th column is assumed to contain an integer
    identifier of the instrument.
    Or list, in which case each element will be one different filename
    containing three columns each.
    """
    if isinstance(datafile, list):
        data = np.empty((0, 5))
        obs = np.empty((0, ))
        for i, df in enumerate(datafile):
            d = np.loadtxt(df, usecols=range(5), skiprows=skip, ndmin=2)
            data = np.append(data, d, axis=0)
            obs = np.append(obs, (i + 1) * np.ones((d.shape[0])))
        return data, obs
    else:
        raise NotImplementedError
        # data = np.loadtxt(datafile, usecols=range(5), skiprows=skip)
        # obs = np.loadtxt(datafile, usecols=(3, ), skiprows=skip, dtype=int)
        # uobs = np.unique(obs)
        # if uobs.min() > 0:
        #     uobs -= uobs.min()
        # return data


@contextlib.contextmanager
def chdir(dir):
    """ A simple context manager to switch directories temporarily """
    curdir = os.getcwd()
    try:
        os.chdir(dir)
        yield
    finally:
        os.chdir(curdir)


def show_tips():
    """ Show a few tips on how to use kima (but only sometimes) """
    tips = (
        "Press Ctrl+C in any of kima's plots to copy the figure.",
        "Run 'kima-showresults all' to plot every figure.",
        "Use the 'kima-template' script to create a new bare-bones directory.")
    if np.random.rand() < 0.2:  # only sometimes, otherwise it's annoying :)
        tip = np.random.choice(tips)
        print('[kima TIP] ' + tip)


def _show_kima_setup():
    def isnotebook():
        try:
            shell = get_ipython().__class__.__name__
            return shell == 'ZMQInteractiveShell'  # notebook or qtconsole
        except NameError:
            return False
    if isnotebook():
        from pygments import highlight
        from pygments.lexers import CppLexer
        from pygments.formatters import HtmlFormatter
        import IPython

        with open('kima_setup.cpp') as f:
            code = f.read()

        formatter = HtmlFormatter()
        return IPython.display.HTML(
            '<style type="text/css">{}</style>{}'.format(
                formatter.get_style_defs('.highlight'),
                highlight(code, CppLexer(), formatter)))
    else:
        with open('kima_setup.cpp') as f:
            print(f.read())


def rms(array):
    """ Root mean square of array """
    return np.sqrt(np.sum(array**2) / array.size)


def wrms(array, weights):
    """ Weighted root mean square of array, given weights """
    mu = np.average(array, weights=weights)
    return np.sqrt(np.sum(weights * (array - mu)**2) / np.sum(weights))


def apply_argsort(arr1, arr2, axis=-1):
    """
    Apply `arr1`.argsort() on `arr2`, along `axis`.
    """
    # check matching shapes
    assert arr1.shape == arr2.shape, "Shapes don't match!"

    i = list(np.ogrid[[slice(x) for x in arr1.shape]])
    i[axis] = arr1.argsort(axis)
    return arr2[i]


def percentile68_ranges(a, min=None, max=None):
    if min is None and max is None:
        mask = np.ones_like(a, dtype=bool)
    elif min is None:
        mask = a < max
    elif max is None:
        mask = a > min
    else:
        mask = (a > min) & (a < max)
    lp, median, up = np.percentile(a[mask], [16, 50, 84])
    return (median, up - median, median - lp)


def percentile68_ranges_latex(a, min=None, max=None):
    median, plus, minus = percentile68_ranges(a, min, max)
    return '$' + urepr.core.uformatul(median, plus, minus, 'L') + '$'


def percentile_ranges(a, percentile=68, min=None, max=None):
    if min is None and max is None:
        mask = np.ones_like(a, dtype=bool)
    elif min is None:
        mask = a < max
    elif max is None:
        mask = a > min
    else:
        mask = (a > min) & (a < max)
    half = percentile / 2
    lp, median, up = np.percentile(a[mask], [50 - half, 50, 50 + half])
    return (median, up - median, median - lp)


def percentile_ranges_latex(a, percentile, min=None, max=None):
    median, plus, minus = percentile_ranges(a, percentile, min, max)
    return '$' + urepr.core.uformatul(median, plus, minus, 'L') + '$'


def clipped_mean(arr, min, max):
    """ Mean of `arr` between `min` and `max` """
    mask = (arr > min) & (arr < max)
    return np.mean(arr[mask])


def clipped_std(arr, min, max):
    """ Standard deviation of `arr` between `min` and `max` """
    mask = (arr > min) & (arr < max)
    return np.std(arr[mask])



# def get_planet_mass_latex(P, K, e, units='earth', **kwargs):
#     """ A simple wrapper around `get_planet_mass` to provide LaTeX strings

#     Args:
#         P : Same argument as for `get_planet_mass`
#         K : Same argument as for `get_planet_mass`
#         e : Same argument as for `get_planet_mass`
#         units (str, optional): Planet mass units, either 'earth' or 'jup'
#         **kwargs : Provided directly to `get_planet_mass`

#     Returns:
#         s (str): LaTeX string with planet mass and possibly uncertainty
#     """
#     out = get_planet_mass(P, K, e, **kwargs)

#     if isinstance(P, float):
#         if units.lower() == 'earth':
#             return '$%f$' % out[1]
#         else:
#             return '$%f$' % out[0]
#     else:
#         if units.lower() == 'earth':
#             return percentile68_ranges_latex(out[2] * mjup2mearth)
#         else:
#             return percentile68_ranges_latex(out[2])


# def get_planet_semimajor_axis_latex(P, K, **kwargs):
#     """
#     A simple wrapper around `get_planet_semimajor_axis` to provide LaTeX strings

#     Args:
#         P : Same argument as for `get_planet_semimajor_axis`
#         K : Same argument as for `get_planet_semimajor_axis`
#         **kwargs : Provided directly to `get_planet_semimajor_axis`

#     Returns:
#         s (str): LaTeX string with planet semi-major axis
#     """
#     out = get_planet_semimajor_axis(P, K, **kwargs)
#     if isinstance(P, float):
#         return '$%f$' % out
#     else:
#         return '$%f$' % out[0]


def get_planet_teq(Tstar: float = 5777, Rstar: float = 1, a: float = 1,
                   A: float = 0, f: float = 1):
    """ Calculate the planet's equlibrium temperature

    Args:
        Tstar (float, optional): Stellar effective temperature [K]. Defaults to 5777.
        Rstar (float, optional): Stellar radius [Rsun]. Defaults to 1.0.
        a (float, optional): Planet semi-major axis [AU]. Defaults to 1.0.
        A (float, optional): Bond albedo. Defaults to 0.
        f (float, optional): Redistribution factor. Defaults to 1.

    Returns:
        Teq (float): Planet equiolibrium temperature [K]
    """
    a = a * AU.to('km')
    Rstar = 1 * R_sun.to('km')
    Teq = Tstar * (f * (1 - A))**(1 / 4) * (Rstar / (2 * a))**(1 / 2)
    return Teq


def get_transit_probability(Rstar=1.0, a=1.0):
    """
    Transit probability, simple. Eq 6.1 in Exoplanet Handbook
    Rstar: stellar radius [Rsun]
    a: semi-major axis [au]
    """
    return Rstar / (a * AU).to(R_sun).value


def lighten_color(color, amount=0.5):
    """
    Lightens the given `color` by multiplying (1-luminosity) by the given
    `amount`. Input can be a matplotlib color string, hex string, or RGB tuple.

    Examples:
    >>> lighten_color('g', 0.3)
    >>> lighten_color('#F034A3', 0.6)
    >>> lighten_color((.3, .55, .1), 0.5)
    """
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except Exception:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


class Fixed:
    """ Not a real distribution, just a parameter fixed at a given `value` """
    def __init__(self, value: float):
        self.value = value

    def rvs(self) -> float:
        """ Random sample (always returns `value`)"""
        return self.value

    def pdf(self, x: float) -> float:
        """ Probability density function: 1.0 if x==value, otherwise 0.0"""
        return 1.0 if x == self.value else 0.0

    def logpdf(self, x: float) -> float:
        """ Logarithm of the probability density function """
        return 0.0 if x == self.value else -np.inf


class ZeroDist:
    """ A dummy probability distribution which always returns 0.0 """
    def __init__(self):
        pass

    def rvs(self) -> float:
        return 0.0

    def pdf(self, x: float) -> float:
        return 0.0

    def logpdf(self, x: float) -> float:
        return 0.0


def _prior_to_dist():
    """ Convert a prior name to a prior object """
    d = {
        'Uniform': stats.uniform,
        'LogUniform': stats.reciprocal,
        'ModifiedLogUniform': ModifiedLogUniform,
        'Gaussian': stats.norm,
        'TruncatedGaussian': stats.truncnorm,
        'Exponential': stats.expon,
        'Kumaraswamy': kumaraswamy,
        'Laplace': stats.laplace,
        'Cauchy': stats.cauchy,
        'InvGamma': lambda shape, scale: stats.invgamma(shape, scale=scale),
        'Fixed': Fixed,
    }
    return d


def _get_prior_parts(prior):
    if not isinstance(prior, str):
        raise ValueError('`prior` should be a string, got', prior)

    try:
        inparens = re.search(r'\((.*?)\)', prior).group(1)
    except AttributeError:
        # print('Cannot decode "%s", seems badly formed' % prior)
        return '', '', prior
    try:
        truncs = re.search(r'\[(.*?)\]', prior).group(1)
    except AttributeError:
        truncs = ''

    name = prior[:prior.find('(')]

    return inparens, truncs, name


def find_prior_limits(prior):
    """
    Find lower and upper limits of a prior from the kima_model_setup.txt file.
    """
    inparens, truncs, name = _get_prior_parts(prior)
    # print(inparens, truncs, name)

    if 'Truncated' in name:
        return tuple(float(v) for v in truncs.split(','))

    if name == 'ModifiedLogUniform':
        return (0.0, float(inparens.split(';')[1]))

    if name == 'Uniform':
        v1, v2 = inparens.split(';')
        return (float(v1), float(v2) - float(v1))

    if name == 'LogUniform':
        return tuple(float(v) for v in inparens.split(';'))

    if name == 'Gaussian':
        return (-np.inf, np.inf)

    if name == 'InvGamma':
        return (0, np.inf)

    if name == 'Kumaraswamy':
        return (0.0, 1.0)


def find_prior_parameters(prior):
    """ Find parameters of a prior from the kima_model_setup.txt file. """
    inparens, _, name = _get_prior_parts(prior)

    truncated = False
    if 'Truncated' in name:
        inparens = ';'.join(inparens.split(';')[:-1])
        name = name.replace('Truncated', '')
        truncated = True

    twopars = ('LogUniform', 'ModifiedLogUniform', 'Gaussian', 'Kumaraswamy',
               'Cauchy', 'InvGamma')

    if name in twopars:
        r = [float(v) for v in inparens.split(';')]
    elif name == 'Uniform':
        v1, v2 = inparens.split(';')
        r = [float(v1), float(v2) - float(v1)]
    elif name in ('Exponential', 'Fixed'):
        r = [float(inparens), ]
    else:
        r = [
            np.nan,
        ]

    if truncated:
        r.append('T')

    return tuple(r)


def get_prior(prior):
    """ Return a scipt.stats-like prior from a kima_model_setup.txt prior """
    _, _, name = _get_prior_parts(prior)
    pars = find_prior_parameters(prior)

    if 'T' in pars:
        a, b = find_prior_limits(prior)
        loc, scale, _ = pars
        a, b = (a - loc) / scale, (b - loc) / scale

    try:
        d = _prior_to_dist()
        if 'T' in pars:
            return d[name](a=a, b=b, loc=loc, scale=scale)
        else:
            return d[name](*pars)
    except KeyError:
        return None


def hyperprior_samples(size):
    pi, log, rng = np.pi, np.log, np.random
    U = rng.rand(size)
    muP = 5.901 + np.tan(pi * (0.97 * rng.rand(size) - 0.485))
    wP = 0.1 + 2.9 * rng.rand(size)
    P = np.where(U < 0.5, muP + wP * log(2. * U), muP - wP * log(2. - 2. * U))
    return np.exp(P)


def priors_latex(results, title='Priors', label='tab:1'):
    priors = results.priors

    tab1 = rf"""\\begin{{table}}
    \r\caption{{{title}}}
    \r\label{{{label}}}
    \r\centering
    \r\\begin{{tabular}}{{l l c}}
    \r  \hline\hline
    \r  Parameter & Units & Prior \\\\
    \r  \hline
    """

    tab2 = r"""  \hline
    \r\end{tabular}
    \r\end{table}
    """

    print(tab1)

    parameter_names = dict(
        Vprior=r'$v_{\rm sys}$',
        C2prior=r'$$',
        Jprior=r'$j^{\rm RV}$',
        J2prior=r'$j^{\rm FWHM}$',
        #
        Pprior='$P$',
        Kprior='$K$',
        eprior='$e$',
        wprior=r'$\omega$',
        phiprior='$M_0$',
        #
        eta1_prior=r'$\eta_1$',
        eta1_1_prior=r'$\eta_1$ RV',
        eta1_2_prior=r'$\eta_1$ FWHM',
        eta2_1_prior=r'$\eta_2$',
        eta3_1_prior=r'$\eta_3$',
        eta4_1_prior=r'$\eta_4$'
    )

    units = dict(
        Vprior=r'\ms',
        C2prior=r'\ms',
        Jprior=r'\ms',
        J2prior=r'\ms',
        #
        Pprior='days',
        Kprior=r'\ms',
        #
        eta1_prior=r'\ms',
        eta1_1_prior=r'\ms',
        eta1_2_prior=r'\ms',
        eta2_1_prior='days',
        eta3_1_prior='days',
    )

    dist_symbols = {
        'uniform': r'$\mathcal{U}$',
        'reciprocal': r'$\mathcal{L}\mathcal{U}$',
        'ModifiedLogUniform': r'$\mathcal{M}\mathcal{L}\mathcal{U}$',
    }
    translation = {
        np.pi: r'$\pi$',
        2 * np.pi: r'$2\pi$',
        results.y.min(): r'$\min v$',
        results.y.max(): r'$\max v$',
        results.t.ptp(): r'$\Delta t$',
        results.y.ptp(): r'$\Delta RV$',
        -results.y.ptp(): r'$-\Delta RV$'
    }
    if results.model == 'RVFWHMmodel':
        translation.update({
            results.y2.min(): r'$\min$ FWHM',
            results.y2.max(): r'$\max$ FWHM',
            results.y2.ptp(): r'$\Delta$ FWHM',
            -results.y2.ptp(): r'$-\Delta$ FWHM'
        })

    def translate_value(value):
        for v, name in translation.items():
            if np.isclose(value, v):
                return name
        return str(value).replace('.0', '')

    for k, prior in priors.items():
        try:
            name = prior.dist.name
        except AttributeError:
            name = prior.__class__.__name__

        try:
            v1, v2 = prior.support()
            v1 = translate_value(v1)
            v2 = translate_value(v2)
            value = ', '.join([v1, v2])
        except AttributeError:
            v = translate_value(prior.val)
            value = v + ' (fixed)'

        s = '  '\
            f'{parameter_names.get(k, k)} & ' \
            f'{units.get(k, "")} & ' \
            f'{dist_symbols.get(name, "")} ({value}) \\\\'
        print(s)
    print()

    print(tab2)


def get_star_name(data_file):
    """ Find star name (usually works for approx. standard filenames) """
    bn = os.path.basename(data_file)
    try:
        pattern = '|'.join([
            r'HD\d+',
            r'HIP\d+',
            r'HR\d+',
            r'BD-\d+',
            r'CD-\d+',
            r'NGC\d+No\d+',
            r'GJ[\d.]+',
            r'Gl\d+',
            r'Proxima',
            r'Barnard',
            r'Ross128',
        ])
        return re.findall(pattern, bn)[0]

    except IndexError:  # didn't find correct name
        if bn.endswith('_harps.rdb'):
            return bn[:-10]

    return 'unknown'


def get_instrument_name(data_file):
    """ Find instrument name """
    bn = os.path.basename(data_file)
    try:
        pattern = '|'.join([
            # 'ESPRESSO',
            # 'ESPRESSO*[\d+]*',
            r'ESPRESSO*[\d+_\w+]*',
            r'HARPS[^\W_]*[\d+]*',
            r'HIRES',
            r'APF',
            r'CORALIE',
            r'HJS',
            r'ELODIE',
            r'KECK',
            r'HET',
            r'LICK',
            r'HRS',
            r'SOPHIE'
        ])
        return re.findall(pattern, bn, re.IGNORECASE)[0]

    except IndexError:
        try:
            return re.findall(pattern, bn.upper())[0]
        except IndexError:
            try:  # at the very least, try removing the file type
                return os.path.splitext(bn)[0]
            except IndexError:  # couldn't do anything, just try basename
                return bn


def read_big_file(filename):
    # return np.genfromtxt(filename)
    try:
        import pandas as pd
        names = open(filename).readline().strip().replace('#', '').split()
        data = pd.read_csv(filename,
                           delim_whitespace=True,
                           comment='#',
                           names=names,
                           dtype=np.float).values

        # pandas.read_csv has problems with 1-line files
        if data.shape[0] == 0:
            data = np.genfromtxt(filename)
        return data

    except ImportError:  # no pandas, use np.genfromtxt
        return np.genfromtxt(filename)


# covert dates to/from Julian days and MJD
# originally from https://gist.github.com/jiffyclub/1294443
# author: Matt Davis (http://github.com/jiffyclub)


def mjd_to_jd(mjd):
    """ Convert Modified Julian Day to Julian Day.

    Args:
        mjd (float): Modified Julian Day

    Returns
        jd (float): Julian Day
    """
    return mjd + 2400000.5


def jd_to_mjd(jd):
    """ Convert Julian Day to Modified Julian Day

    Args:
        jd (float): Julian Day

    Returns
        mjd (float): Modified Julian Day
    """
    return jd - 2400000.5


def date_to_jd(year, month, day):
    """ Convert a date (year, month, day) to Julian Day.

    Algorithm from 'Practical Astronomy with your Calculator or Spreadsheet',
        4th ed., Duffet-Smith and Zwart, 2011.

    Args:
        year (int): 
            Year as integer. Years preceding 1 A.D. should be 0 or negative. The
            year before 1 A.D. is 0, 10 B.C. is year -9.
        month (int): Month as integer, Jan = 1, Feb. = 2, etc.
        day (float): Day, may contain fractional part.

    Returns:
        jd (float): Julian Day

    Examples:
        Convert 6 a.m., February 17, 1985 to Julian Day
        >>> date_to_jd(1985, 2, 17.25)
        2446113.75
    """
    if month == 1 or month == 2:
        yearp = year - 1
        monthp = month + 12
    else:
        yearp = year
        monthp = month

    # this checks where we are in relation to October 15, 1582, the beginning
    # of the Gregorian calendar.
    if (year < 1582 or (year == 1582 and month < 10)
            or (year == 1582 and month == 10 and day < 15)):
        # before start of Gregorian calendar
        B = 0
    else:
        # after start of Gregorian calendar
        A = math.trunc(yearp / 100.)
        B = 2 - A + math.trunc(A / 4.)

    if yearp < 0:
        C = math.trunc((365.25 * yearp) - 0.75)
    else:
        C = math.trunc(365.25 * yearp)

    D = math.trunc(30.6001 * (monthp + 1))

    jd = B + C + D + day + 1720994.5

    return jd


def jd_to_date(jd):
    """ Convert Julian Day to date.

    Algorithm from 'Practical Astronomy with your Calculator or Spreadsheet',
        4th ed., Duffet-Smith and Zwart, 2011.

    Args:
        jd (float): Julian Day

    Returns:
        year (int): 
            Year as integer. Years preceding 1 A.D. should be 0 or negative. The
            year before 1 A.D. is 0, 10 B.C. is year -9.
        month (int): Month as integer, Jan = 1, Feb. = 2, etc.
        day (float): Day, may contain fractional part.

    Examples:
        Convert Julian Day 2446113.75 to year, month, and day.
        >>> jd_to_date(2446113.75)
        (1985, 2, 17.25)
    """
    jd = jd + 0.5

    frac, integ = math.modf(jd)
    integ = int(integ)

    A = math.trunc((integ - 1867216.25) / 36524.25)

    if integ > 2299160:
        B = integ + 1 + A - math.trunc(A / 4.)
    else:
        B = integ

    C = B + 1524

    D = math.trunc((C - 122.1) / 365.25)

    E = math.trunc(365.25 * D)

    G = math.trunc((C - E) / 30.6001)

    day = C - E + frac - math.trunc(30.6001 * G)

    if G < 13.5:
        month = G - 1
    else:
        month = G - 13

    if month > 2.5:
        year = D - 4716
    else:
        year = D - 4715

    return year, month, day


def hmsm_to_days(hour=0, min=0, sec=0, micro=0):
    """
    Convert hours, minutes, seconds, and microseconds to fractional days.

    Args:
        hour (int, optional): Hour number. Defaults to 0.
        min (int, optional): Minute number. Defaults to 0.
        sec (int, optional): Second number. Defaults to 0.
        micro (int, optional): Microsecond number. Defaults to 0.

    Returns:
        days (float): Fractional days

    Examples:
        >>> hmsm_to_days(hour=6)
        0.25
    """
    days = sec + (micro / 1.e6)

    days = min + (days / 60.)

    days = hour + (days / 60.)

    return days / 24.


def days_to_hmsm(days):
    """
    Convert fractional days to hours, minutes, seconds, and microseconds.
    Precision beyond microseconds is rounded to the nearest microsecond.

    Args:
        days (float): A fractional number of days. Must be less than 1.

    Returns:
        hour (int): Hour number
        min (int): Minute number
        sec (int): Second number
        micro (int): Microsecond number

    Raises:
        ValueError: If `days` is >= 1.

    Examples:
        >>> days_to_hmsm(0.1)
        (2, 24, 0, 0)
    """
    hours = days * 24.
    hours, hour = math.modf(hours)

    mins = hours * 60.
    mins, min = math.modf(mins)

    secs = mins * 60.
    secs, sec = math.modf(secs)

    micro = round(secs * 1.e6)

    return int(hour), int(min), int(sec), int(micro)


def datetime_to_jd(date: datetime):
    """ Convert a `datetime.datetime` object to Julian day.

    Args:
        date (datetime): the date to be converted to Julian day

    Returns:
        jd (float): Julian day

    Examples:
        >>> d = datetime.datetime(1985, 2, 17, 6)
        >>> d
        datetime.datetime(1985, 2, 17, 6, 0)
        >>> jdutil.datetime_to_jd(d)
        2446113.75
    """
    days = date.day + hmsm_to_days(date.hour, date.minute, date.second,
                                   date.microsecond)

    return date_to_jd(date.year, date.month, days)


def styleit(func):
    from matplotlib.pyplot import style
    here = os.path.dirname(__file__)

    def wrapper(*args, **kwargs):
        with style.context([os.path.join(here, 'simple.mplstyle')]):
            return func(*args, **kwargs)

    return wrapper



from dataclasses import dataclass
from functools import partial

import numpy as np
# from numpy import exp, sin, cos, sqrt, abs

from jax.config import config
config.update("jax_enable_x64", True)

import jax.numpy as jnp
from jax.numpy import exp, sin, cos, sqrt
from jax import jit, vmap, jacfwd

from scipy.spatial.distance import squareform, pdist, cdist
from scipy.linalg import cholesky, cho_solve, solve_triangular


__all__ = [
    'QPkernel',
    'QPCkernel',
    'QPMatern32kernel',
    'QPMatern52kernel',
    'QPRQkernel',
    'SqExpkernel',
    'GP',
]


π = np.pi


class Kernel:

    @property
    def pars(self):
        return np.array([getattr(self, n) for n in self.names])

    @pars.setter
    def pars(self, newpars):
        newpars = np.atleast_1d(newpars).astype(float)
        if newpars.size != self.npars:
            name = self.__class__.__name__
            raise ValueError(
                f'{name} expects {self.npars} parameters, got {newpars.size}')
        for n, p in zip(self.names, newpars):
            setattr(self, n, p)

    # @partial(jit, static_argnums=(0,))
    # def metric(self, x1, x2=None):
    #     if x2 is None:
    #         return abs(x1[None, :] - x1[:, None])
    #     else:
    #         if x1.ndim == 1:
    #             x1 = x1.reshape(-1, 1)
    #         if x2.ndim == 1:
    #             x2 = x2.reshape(-1, 1)
    #         return abs(x1 - x2.T)

    @partial(jit, static_argnums=(0,))
    def __call__(self, xi, xj=None):
        if xj is None:
            return vmap(lambda x: vmap(lambda y: self.eval(x, y))(xi))(xi)
        else:
            return vmap(lambda x: vmap(lambda y: self.eval(x, y))(xi))(xj).T

    @partial(jit, static_argnums=(0,))
    def gradient_wrt_x1(self, xi, xj=None):
        i = jnp.arange(xi.size)
        return jacfwd(self.__call__, argnums=0)(xi, xj)[i, :, i]



@dataclass(eq=True, unsafe_hash=True)
class COSkernel(Kernel):
    """ The cosine kernel """
    η1: float
    P: float
    names = ['η1', 'P']
    npars = 2

    @partial(jit, static_argnums=(0,))
    def eval(self, xi, xj):
        r = xi - xj
        K = self.η1**2 * cos(2 * π * r / self.P)
        return K


@dataclass(eq=True, unsafe_hash=True)
class DopplerCOSkernel(Kernel):
    """ The cosine kernel """
    η1: float
    P: float
    RV: float
    names = ['η1', 'P', 'RV']
    npars = 3

    @partial(jit, static_argnums=(0,))
    def eval(self, xi, xj):
        α = 1 + self.RV / 299792458.0
        r = α * (xi - xj)
        K = self.η1**2 * cos(2 * π * r / self.P)
        return K


@dataclass(eq=True, unsafe_hash=True)
class RBFkernel(Kernel):
    """ The radial basis function (exponential squared) kernel """
    η1: float
    η2: float
    names = ['η1', 'η2']
    npars = 2

    @partial(jit, static_argnums=(0,))
    def eval(self, xi, xj):
        r = xi - xj
        sr = r / self.η2
        K = self.η1**2 * exp(-0.5 * sr**2)
        return K


@dataclass(eq=True, unsafe_hash=True)
class QPkernel(Kernel):
    """ The quasi-periodic kernel """
    η1: float
    η2: float
    η3: float
    η4: float
    names = ['η1', 'η2', 'η3', 'η4']
    npars = 4

    @partial(jit, static_argnums=(0,))
    def eval(self, xi, xj):
        r = xi - xj
        sr = r / self.η2
        pr = π * r / self.η3
        K = self.η1**2 * exp(-0.5 * sr**2 - 2 * (sin(pr) / self.η4)**2)
        return K


@dataclass(eq=True, unsafe_hash=True)
class QPCkernel(Kernel):
    """ The quasi-periodic-cosine kernel [Perger+2020]"""
    η1: float
    η2: float
    η3: float
    η4: float
    η5: float
    names = ['η1', 'η2', 'η3', 'η4', 'η5']
    npars = 5

    def eval(self, xi, xj):
        r = xi - xj
        sr = r / self.η2
        pr = π * r / self.η3
        η1 = self.η1**2
        η5 = self.η5**2
        K = exp(-0.5 * sr**2) * (η1 * exp(-2 * (sin(pr) / self.η4)**2) +
                                 η5 * cos(4 * pr))
        return K


@dataclass(eq=True, unsafe_hash=True)
class QPpCkernel(Kernel):
    """ The quasi-periodic *plus* cosine kernel """
    η1: float
    η2: float
    η3: float
    η4: float
    η5: float
    η6: float
    names = ['η1', 'η2', 'η3', 'η4', 'η5', 'η6']
    npars = 6

    @partial(jit, static_argnums=(0,))
    def eval(self, xi, xj):
        r = xi - xj
        sr = r / self.η2
        pr = π * r / self.η3
        K = self.η1**2 * exp(-0.5 * sr**2 - 2 * (sin(pr) / self.η4)**2) \
            + self.η5**2 * cos(2 * π * r / self.η6)
        return K


@dataclass(eq=True, unsafe_hash=True)
class PERkernel(Kernel):
    """ The periodic kernel [Rasmussen & Williams 2006]"""
    η1: float
    η3: float
    η4: float
    names = ['η1', 'η3', 'η4']
    npars = 3

    def eval(self, xi, xj):
        r = xi - xj
        pr = π * r / self.η3
        K = self.η1**2 * exp(- 2 * (sin(pr) / self.η4)**2)
        return K


class QPMatern32kernel():
    """ The quasi-periodic kernel based on the Matern 3/2 """
    def __init__(self, eta1, eta2, eta3, eta4):
        self.eta1 = eta1
        self.eta2 = eta2
        self.eta3 = eta3
        self.eta4 = eta4

    def setpars(self, eta1=None, eta2=None, eta3=None, eta4=None):
        self.eta1 = eta1 if eta1 else self.eta1
        self.eta2 = eta2 if eta2 else self.eta2
        self.eta3 = eta3 if eta3 else self.eta3
        self.eta4 = eta4 if eta4 else self.eta4

    def __call__(self, x1, x2=None):
        if x1.ndim == 1:
            x1 = x1.reshape(-1, 1)

        if x2 is None:
            dists = squareform(pdist(x1, metric='euclidean'))
        else:
            if x2.ndim == 1:
                x2 = x2.reshape(-1, 1)
            dists = cdist(x1, x2, metric='euclidean')

        argexp = dists / self.eta2
        argsin = np.pi * dists / self.eta3
        K = self.eta1**2 \
            * exp(-0.5 * argexp**2) \
            * (1 + sqrt(3) * 2 * np.abs(sin(argsin) / self.eta4)) \
            * exp(-sqrt(3) * 2 * np.abs(sin(argsin) / self.eta4))

        return K


class QPMatern52kernel():
    """ The quasi-periodic kernel based on the Matern 5/2 """
    def __init__(self, eta1, eta2, eta3, eta4):
        self.eta1 = eta1
        self.eta2 = eta2
        self.eta3 = eta3
        self.eta4 = eta4

    def setpars(self, eta1=None, eta2=None, eta3=None, eta4=None):
        self.eta1 = eta1 if eta1 else self.eta1
        self.eta2 = eta2 if eta2 else self.eta2
        self.eta3 = eta3 if eta3 else self.eta3
        self.eta4 = eta4 if eta4 else self.eta4

    def __call__(self, x1, x2=None):
        if x1.ndim == 1:
            x1 = x1.reshape(-1, 1)

        if x2 is None:
            dists = squareform(pdist(x1, metric='euclidean'))
        else:
            if x2.ndim == 1:
                x2 = x2.reshape(-1, 1)
            dists = cdist(x1, x2, metric='euclidean')

        argexp = dists / self.eta2
        argsin = np.pi * dists / self.eta3
        s = 2 * np.abs(sin(argsin))

        K = self.eta1**2 \
            * exp(-0.5 * argexp**2) \
            * (1 + sqrt(5) * s / self.eta4 + 5 * s**2 / (3 * self.eta4**2)) \
            * exp(-sqrt(5) * s / self.eta4)

        return K


class QPRQkernel():
    """ The quasi-periodic kernel based on the Rational Quadratic """
    def __init__(self, eta1, eta2, eta3, eta4, alpha):
        self.eta1 = eta1
        self.eta2 = eta2
        self.eta3 = eta3
        self.eta4 = eta4
        self.alpha = alpha

    def setpars(self, eta1=None, eta2=None, eta3=None, eta4=None, alpha=None):
        self.eta1 = eta1 if eta1 else self.eta1
        self.eta2 = eta2 if eta2 else self.eta2
        self.eta3 = eta3 if eta3 else self.eta3
        self.eta4 = eta4 if eta4 else self.eta4
        self.alpha = alpha if alpha else self.alpha

    def __call__(self, x1, x2=None):
        if x1.ndim == 1:
            x1 = x1.reshape(-1, 1)

        if x2 is None:
            dists = squareform(pdist(x1, metric='euclidean'))
        else:
            if x2.ndim == 1:
                x2 = x2.reshape(-1, 1)
            dists = cdist(x1, x2, metric='euclidean')

        argexp = dists / self.eta2
        s = np.abs(sin(np.pi * dists / self.eta3))

        K = self.eta1**2 * exp(-0.5 * argexp**2) \
                * (1 + 2*s**2/(self.alpha*self.eta4**2))**(-self.alpha)

        return K


class SqExpkernel():
    """ The quasi-periodic kernel """
    def __init__(self, eta1, eta2):
        self.eta1 = eta1
        self.eta2 = eta2

    def setpars(self, eta1=None, eta2=None):
        self.eta1 = eta1 if eta1 else self.eta1
        self.eta2 = eta2 if eta2 else self.eta2

    def __call__(self, x1, x2=None):
        if x1.ndim == 1:
            x1 = x1.reshape(-1, 1)

        if x2 is None:
            dists = squareform(pdist(x1, metric='euclidean'))
            argexp = dists / self.eta2
            K = self.eta1**2 * np.exp(-0.5 * argexp**2)
        else:
            if x2.ndim == 1:
                x2 = x2.reshape(-1, 1)
            dists = cdist(x1, x2, metric='euclidean')
            argexp = dists / self.eta2
            K = self.eta1**2 * np.exp(-0.5 * argexp**2)

        return K


class GP():
    """ A simple Gaussian process class for regression problems """
    def __init__(self, kernel, t, yerr=None, white_noise=1.25e-12):
        """
        kernel : an instance of `QPkernel`
        t : the independent variable where the GP is "trained"
        (opt) yerr : array of observational uncertainties
        (opt) white_noise : value/array added to the diagonal of the kernel matrix
        """
        self.kernel = kernel
        self.t = t
        self.yerr = np.zeros_like(t) if yerr is None else yerr
        self.white_noise = white_noise
        self._C1 = -0.5 * t.size * np.log(2 * π)

    def _cov(self, xi, xj=None, add2diag=True):
        """ Calculate the kernel matrix evaluated at inputs xi (and xj) """
        K = self.kernel(xi, xj)
        if add2diag:
            K = K + jnp.diag(self.yerr**2)
            K = K + jnp.eye(self.t.size) * self.white_noise**2
        return K

    def log_likelihood(self, y):
        """
        The log marginal likelihood of observations y under the GP model
        """
        self.K = self._cov(self.t)
        self.L_ = cholesky(self.K, lower=True)
        self.alpha_ = cho_solve((self.L_, True), y)
        log_determinant = 2 * np.log(np.diag(self.L_)).sum()
        return self._C1 - 0.5 * log_determinant - 0.5 * y.dot(self.alpha_)

    def sample(self, t=None, size=1):
        """ Draw samples from the GP prior distribution;
        Output shape: (t.size, size) 
        """
        if t is None:
            K = self._cov(self.t)
        else:
            K = self._cov(t, add2diag=False)
        return multivariate_gaussian_samples(K, size).T

    def predict(self, y, t=None, return_std=False, return_cov=False):
        """
        Conditional predictive distribution of the GP model given observations
        y, evaluated at coordinates t.
        """
        if t is None:
            t = self.t

        self.K = self._cov(self.t)
        self.L_ = cholesky(self.K, lower=True)
        self.alpha_ = cho_solve((self.L_, True), y)
        K_trans = self.kernel(t, self.t)
        mean = K_trans.dot(self.alpha_)

        if return_cov:
            v = cho_solve((self.L_, True), K_trans.T)  # Line 5
            cov = self.kernel(t) - K_trans.dot(v)  # Line 6
            return mean, cov
        elif return_std:
            # compute inverse K_inv of K based on its Cholesky
            # decomposition L and its inverse L_inv
            L_inv = solve_triangular(self.L_.T, np.eye(self.L_.shape[0]))
            self._K_inv = L_inv.dot(L_inv.T)

            # Compute variance of predictive distribution
            var = np.diag(self.kernel(t))
            var = var - np.einsum("ij,ij->i",
                                  np.dot(K_trans, self._K_inv), K_trans)

            # Check if any of the variances is negative because of
            # numerical issues. If yes: set the variance to 0.
            var_negative = var < 0
            if np.any(var_negative):
                var[var_negative] = 0.0
            return mean, np.sqrt(var)
        else:
            return mean

    def derivative(self, y, t=None, return_std=False, return_cov=False):
        """
        First derivative of the conditional predictive distribution of the GP
        model given observations y, evaluated at coordinates t.
        """
        if t is None:
            t = self.t

        self.K = self._cov(self.t)
        self.L_ = cholesky(self.K, lower=True)
        self.alpha_ = cho_solve((self.L_, True), y)

        deriv = self.kernel.gradient_wrt_x1(self.t, t).T @ self.alpha_
        return deriv

        K_trans = self.kernel(t, self.t)
        mean = K_trans.dot(self.alpha_)

        if return_cov:
            v = cho_solve((self.L_, True), K_trans.T)  # Line 5
            cov = self.kernel(t) - K_trans.dot(v)  # Line 6
            return mean, cov
        elif return_std:
            # compute inverse K_inv of K based on its Cholesky
            # decomposition L and its inverse L_inv
            L_inv = solve_triangular(self.L_.T, np.eye(self.L_.shape[0]))
            self._K_inv = L_inv.dot(L_inv.T)

            # Compute variance of predictive distribution
            var = np.diag(self.kernel(t))
            var = var - np.einsum("ij,ij->i",
                                  np.dot(K_trans, self._K_inv), K_trans)

            # Check if any of the variances is negative because of
            # numerical issues. If yes: set the variance to 0.
            var_negative = var < 0
            if np.any(var_negative):
                var[var_negative] = 0.0
            return mean, np.sqrt(var)
        else:
            return mean

    def predict_with_hyperpars(self, results, sample, t=None, add_parts=True,
                               return_std=False):
        """ 
        Given the parameters in `sample`, return the GP predictive mean. If `t`
        is None, the prediction is done at the observed times and will contain
        other model components (systemic velocity, trend, instrument offsets) if
        `add_parts` is True. Otherwise, when `t` is given, the prediction is
        made at times `t` and *does not* contain these extra model components.
        """
        ind = results.indices['GPpars']
        GP_pars_sample = sample[ind]
        s = sample[0]

        self.kernel.setpars(*GP_pars_sample)
        self.white_noise = s

        y = results.y.copy()
        # put all data around 0
        y -= sample[-1]
        if results.multi:
            for i in range(1, results.n_instruments):
                y[results.obs == i] -= sample[results.indices['inst_offsets']][i-1]
        # and subtract the trend
        if results.trend:
            y -= sample[results.indices['trend']] * (results.t - results.tmiddle)

        if t is not None:
            pred = self.predict(y, t, return_cov=False, return_std=return_std)
            return pred

        mu = self.predict(y, results.t, return_cov=False,
                          return_std=return_std)

        if add_parts:
            # add the trend back
            mu += sample[results.indices['trend']] * (results.t - results.tmiddle)
            # add vsys and instrument offsets back
            mu += sample[-1]
            if results.multi:
                for i in range(1, results.n_instruments):
                    mu[results.obs == i] += sample[results.indices['inst_offsets']][i-1]

        return mu

    def sample_conditional(self, y, t, size=1):
        """ 
        Sample from the conditional predictive distribution of the GP model 
        given observations y, at coordinates t.
        """
        mu, cov = self.predict(y, t, return_cov=True)
        return multivariate_gaussian_samples(cov, size, mean=mu).T

    def sample_from_posterior(self, results, size=1):
        """ 
        Given the posterior sample in `results`, take one sample of the GP
        hyperparameters and the white noise, and return a sample from the GP
        prior given those parameters.
        """
        # choose a random sample
        i = np.random.choice(range(results.posterior_sample.shape[0]))
        sample = results.posterior_sample[i, :]
        ind = results.indices['GPpars']
        GPpars = sample[ind]
        s = sample[0]

        self.kernel.setpars(*GPpars)
        self.white_noise = s
        return self.sample(size=size)

    def sample_with_hyperpars(self, results, sample, size=1):
        """ 
        Given the value of the hyperparameters and the white noise in `sample`, return a 
        sample from the GP prior.
        """
        ind = results.indices['GPpars']
        GPpars = sample[ind]
        # s = sample[0]

        if results.model == 'RVFWHMmodel':
            η1RV, η1FWHM, η2RV, η2FWHM, η3RV, η3FWHM, η4RV, η4FWHM = \
                GPpars[results._GP_par_indices]

            if results.GPkernel == 'standard':
                results.GP1.kernel.pars = np.array([η1RV, η2RV, η3RV, η4RV])
                results.GP2.kernel.pars = np.array([η1FWHM, η2FWHM, η3FWHM, η4FWHM])
        else:
            results.GP.kernel.pars = GPpars

        return self.sample(size=size)

    def sample_conditional_from_posterior(self, results, t, size=1):
        """ 
        Given the posterior sample in `results`, take one sample of the GP
        hyperparameters and the white noise, and return a sample from the GP
        predictive given those parameters.
        """
        i = np.random.choice(range(results.posterior_sample.shape[0]))
        ind = results.indices['GPpars']
        sample = results.posterior_sample[i,:]
        GPpars = sample[ind]
        s = sample[0]

        self.kernel.setpars(*GPpars)
        self.white_noise = s

        # put all data around 0
        y = results.y.copy()
        y -= sample[-1]
        if results.multi:
            for i in range(1, results.n_instruments):
                y[results.obs == i] -= sample[results.indices['inst_offsets']]
        # and subtract the trend
        if results.trend:
            y -= sample[results.indices['trend']] * (t - results.tmiddle)

        mu, cov = self.predict(y, t, return_cov=True)
        return multivariate_gaussian_samples(cov, size, mean=mu).T

    def sample_conditional_with_hyperpars(self, results, sample, t, size=1):
        """ 
        Given the value of the hyperparameters and the white noise in `sample`,
        return a sample from the GP predictive.
        """
        ind = results.indices['GPpars']
        GPpars = sample[ind]
        s = sample[0]

        if results.model == 'RVFWHMmodel':
            η1RV, η1FWHM, η2RV, η2FWHM, η3RV, η3FWHM, η4RV, η4FWHM = \
                GPpars[results._GP_par_indices]

            if results.GPkernel == 'standard':
                results.GP1.kernel.pars = np.array([η1RV, η2RV, η3RV, η4RV])
                results.GP2.kernel.pars = np.array([η1FWHM, η2FWHM, η3FWHM, η4FWHM])
        else:
            results.GP.kernel.pars = GPpars

        if results.model == 'RVFWHMmodel':
            # which GP am I?
            if (results.e == self.yerr).all():
                print('1')
                y = results.y - results.eval_model(sample)[0]
            else:
                print('2')
                y = results.y2 - results.eval_model(sample)[1]
        else:
            y = results.y - results.eval_model(sample)

        mu, cov = self.predict(y, t, return_cov=True)
        return multivariate_gaussian_samples(cov, size, mean=mu).T


from tinygp import GaussianProcess
from tinygp.kernels import Matern32, ExpSquared, ExpSineSquared
from tinygp import transforms


class mixtureGP(GP):
    def __init__(self, kernels, X, dims, yerr=None, white_noise=1.25e-12):
        self.X = X
        self.wn = white_noise

    @partial(jit, static_argnums=(0,))
    def build_gp_predict(self, y, T, params, X, wn):
        qp = ExpSquared(params["scale1"]) \
              * ExpSineSquared(scale=params["period"], gamma=params["gamma"])
        kernel1 = params["amp1"] * transforms.Subspace(0, qp)
        kernel2 = params["amp2"] * transforms.Subspace(1, ExpSquared(scale=params["scale2"]))

        kernel = kernel1 + kernel2
        gp = GaussianProcess(kernel, X, diag=wn)
        return gp.predict(y, T)

    def predict(self, y, T, pars):
        params = {
            'amp1': pars[0],
            'scale1': pars[1],
            'period': pars[2],
            'gamma': pars[3],
            'amp2': pars[4],
            'scale2': pars[5],
        }
        mu = self.build_gp_predict(y, T, params, self.X, self.wn)
        return mu

    def predict_separate(self, y, T, pars):
        params = {
            'amp1': pars[0],
            'scale1': pars[1],
            'period': pars[2],
            'gamma': pars[3],
            'amp2': pars[4],
            'scale2': pars[5],
        }
        gp = self.build_gp(params)
        mu1 = gp.predict(y, T, kernel=gp.kernel.kernel1)
        mu2 = gp.predict(y, T, kernel=gp.kernel.kernel2)
        return mu1, mu2


# class QPkernel_celerite(terms.Term):
#     # This implements a quasi-periodic kernel (QPK) devised by Andrew Collier
#     # Cameron, which mimics a standard QP kernel with a roughness parameter 0.5,
#     # and has zero derivative at the origin: k(tau=0)=amp and k'(tau=0)=0
#     # The kernel defined in the celerite paper (Eq 56 in Foreman-Mackey et al. 2017)
#     # does not satisfy k'(tau=0)=0
#     #
#     # This new QPK has only 3 parameters, here called eta1, eta2, eta3
#     # corresponding to an amplitude, decay timescale and period.
#     """
#     docs
#     """
#     parameter_names = ("η1", "η2", "η3")

#     def get_real_coefficients(self, params):
#         # η1, η2, η3 = params
#         return np.zeros(0, dtype=np.float), np.zeros(0, dtype=np.float)

#     def get_complex_coefficients(self, params):
#         η1, η2, η3 = params
#         wbeat = 1.0 / η2
#         wrot = 2*np.pi/η3
#         amp = η1**2
#         c = wbeat
#         d = wrot
#         x = c/d
#         a = amp/2.
#         b = amp*x/2.
#         e = amp/8.
#         f = amp*x/4.
#         g = amp*(3./8. + 0.001)
#         return (
#             np.reshape([a, e, g], (3,)),
#             np.reshape([b, f, 0], (3,)),
#             np.reshape([c, c, c], (3,)),
#             np.reshape([d, 2*d, 0], (3,)),
#         )

#     @property
#     def J(self):
#         return 6

#     def setpars(self, eta1=None, eta2=None, eta3=None, eta4=None):
#         p = [eta1, eta2, eta3]
#         # if eta1: p.append(eta1)
#         # if eta2: p.append(eta2)
#         # if eta3: p.append(eta3)
#         self.set_parameter_vector(p)


# class GP_celerite():
#     """ A simple Gaussian process class for regression problems, based on celerite """
#     def __init__(self, kernel, t, yerr=None, white_noise=1e-10):
#         """
#         t : the independent variable where the GP is "trained"
#         (opt) yerr : array of observational uncertainties
#         (opt) white_noise : value/array added to the diagonal of the kernel matrix
#         """
#         self.kernel = kernel
#         self.t = t
#         self.yerr = yerr
#         self.white_noise = white_noise
#         self._GP = GPcel(self.kernel)
#         self._GP.compute(self.t, self.yerr)

#     def _cov(self, x1, x2=None, add2diag=True):
#         raise NotImplementedError

#     def log_likelihood(self, y):
#         """ The log marginal likelihood of observations y under the GP model """
#         raise NotImplementedError

#     def sample(self, t=None, size=1):
#         """ Draw samples from the GP prior distribution;
#         Output shape: (t.size, size)
#         """
#         raise NotImplementedError

#     def predict(self, y, t=None, return_std=False, return_cov=False):
#         """
#         Conditional predictive distribution of the GP model
#         given observations y, evaluated at coordinates t.
#         """
#         return self._GP.predict(y, t, return_cov=return_cov, return_var=return_std)


#     def predict_with_hyperpars(self, results, sample, t=None, add_parts=True):
#         """
#         Given the parameters in `sample`, return the GP predictive mean. If `t`
#         is None, the prediction is done at the observed times and will contain
#         other model components (systemic velocity, trend, instrument offsets) if
#         `add_parts` is True. Otherwise, when `t` is given, the prediction is
#         made at times `t` and *does not* contain these extra model components.
#         """
#         raise NotImplementedError

#     def sample_conditional(self, y, t, size=1):
#         """
#         Sample from the conditional predictive distribution of the GP model
#         given observations y, at coordinates t.
#         """
#         raise NotImplementedError


#     def sample_from_posterior(self, results, size=1):
#         """
#         Given the posterior sample in `results`, take one sample of the GP
#         hyperparameters and the white noise, and return a sample from the GP
#         prior given those parameters.
#         """
#         raise NotImplementedError

#     def sample_with_hyperpars(self, results, sample, size=1):
#         """
#         Given the value of the hyperparameters and the white noise in `sample`, return a
#         sample from the GP prior.
#         """
#         raise NotImplementedError


#     def sample_conditional_from_posterior(self, results, t, size=1):
#         """
#         Given the posterior sample in `results`, take one sample of the GP
#         hyperparameters and the white noise, and return a sample from the GP
#         predictive given those parameters.
#         """
#         raise NotImplementedError

#     def sample_conditional_with_hyperpars(self, results, sample, t, size=1):
#         """
#         Given the value of the hyperparameters and the white noise in `sample`,
#         return a sample from the GP predictive.
#         """
#         raise NotImplementedError


def multivariate_gaussian_samples(matrix, N, mean=None):
    """
    Generate samples from a multidimensional Gaussian with a given covariance.

    :param matrix: ``(k, k)``
        The covariance matrix.

    :param N:
        The number of samples to generate.

    :param mean: ``(k,)`` (optional)
        The mean of the Gaussian. Assumed to be zero if not given.

    :returns samples: ``(k,)`` or ``(N, k)``
        Samples from the given multivariate normal.

    """
    if mean is None:
        mean = np.zeros(len(matrix))
    samples = np.random.multivariate_normal(mean, matrix, N)
    if N == 1:
        return samples[0]
    return samples
